"""Module containing functionality for preparing arguments for the "workspace" type spec file template."""

import time
from dataclasses import KW_ONLY, dataclass
from functools import cached_property

from cargo2rpm import rpm
from cargo2rpm.metadata import Metadata
from cargo2rpm.semver import Version

from rust2rpm import __version__, log
from rust2rpm.conf import Patch, Source, TomlConf
from rust2rpm.project import LocalProject
from rust2rpm.sysinfo import Target

from .common import (
    RUST_PACKAGING_DEPS,
    cargo_args_from_flags,
    conf_to_bcond_check,
    conf_to_bcond_check_comments,
    conf_to_bin_rename_commands,
    conf_to_cargo_test_commands,
    conf_to_feature_flags,
    make_rpm_summary,
    min_rust_packaging_dep,
    normalize_spdx_expr,
    spec_file_template,
)
from .meta import (
    InvalidMetadataError,
    Parameters,
    ParametersFedora,
    ParametersMageia,
    ParametersOpenSUSE,
    ParametersPlain,
    Template,
    parameters_as_dict,
)


def _license_is_composite(expr: str) -> bool:
    return (" " in expr and " WITH " not in expr and "(" not in expr and ")" not in expr) or "(" in expr or ")" in expr


@dataclass(frozen=True)
class WorkspaceParameters(Parameters):
    """Collection of parameters for the "workspace" type template that is common for all targets."""

    _: KW_ONLY

    _project: LocalProject
    _conf: TomlConf
    _target: Target

    _make_changelog_entry: bool

    # Convenience properties

    @property
    def _metadata(self) -> Metadata:
        return self._project.metadata

    @cached_property
    def _rpm_license_and_comments(self) -> tuple[str | None, str | None]:
        license_strs = list({package.license for package in self._metadata.packages if package.license})

        if len(license_strs) == 0:
            msg = (
                "None of the crates in this workspace project specify a license in crate "
                "metadata. The 'package.license' property MUST be set in at least one workspace "
                "member, otherwise RPM packaging macros will not work properly. To resolve this "
                "issue, set the 'package.license' property in at least one workspace member to "
                "the correct license expression in SPDX format with a patch."
            )
            raise InvalidMetadataError(msg)

        if len(license_strs) == 1:
            rpm_license_tag, rpm_license_comments = normalize_spdx_expr(license_strs[0]), None

        else:
            license_strs = [
                license_str if not _license_is_composite(license_str) else f"({license_str})"
                for license_str in license_strs
                if license_str is not None
            ]
            license_strs.sort()
            license_str = " AND ".join(license_strs)
            rpm_license_tag, rpm_license_comments = normalize_spdx_expr(license_str), None

        return rpm_license_tag, rpm_license_comments

    @cached_property
    def _buildrequires(self) -> tuple[list[str], list[str]]:
        feature_flags = conf_to_feature_flags(self._conf)

        buildrequires = rpm.workspace_buildrequires(self._metadata, feature_flags, with_dev_deps=False)
        test_requires = set.difference(
            rpm.workspace_buildrequires(self._metadata, feature_flags, with_dev_deps=True),
            rpm.workspace_buildrequires(self._metadata, feature_flags, with_dev_deps=False),
        )

        return sorted(buildrequires), sorted(test_requires)

    # Parameters specific to rust2rpm

    @property
    def rust2rpm_version(self) -> str:
        """Current major version of rust2rpm."""
        return __version__.split(".")[0]

    @property
    def rust2rpm_target(self) -> str:
        """Target platform (fedora, epel8, mageia, opensuse, plain)."""
        return self._target

    @property
    def rust_packaging_dep(self) -> str:
        """Dependency string for RPM Rust packaging tools."""
        return RUST_PACKAGING_DEPS[
            max(
                (
                    24,
                    max(
                        min_rust_packaging_dep(
                            package,
                            self._target,
                            self._project.vendor_tarball,
                            is_bin=package.is_bin(),
                            is_cdylib=package.is_cdylib(),
                            cargo_install_lib=True,
                            cargo_install_bin=True,
                        )
                        for package in self._metadata.packages
                    ),
                ),
            )
        ]

    # Parameters that control compiler flags

    @property
    def build_rustflags_debuginfo(self) -> int | None:
        """Controls the level of debuginfo generated by rustc."""
        return self._conf.package.debuginfo_level

    # Parameters for RPM package metadata

    @property
    def rpm_name(self) -> str:
        """RPM source package Name."""
        return self._project.rpm_name

    @property
    def rpm_version(self) -> str:
        """RPM package Version (translated to RPM format from SemVer)."""
        try:
            return Version.parse(self._project.version).to_rpm()
        except ValueError:
            log.warn(f"Version {self._project.version!r} is not valid according to SemVer.")
            return self._project.version

    @cached_property
    def rpm_summary(self) -> str | None:
        """RPM package summary (derived from package.description value from Cargo.toml)."""
        return make_rpm_summary(self._conf, self._project.main_package)

    @property
    def rpm_description(self) -> str | None:
        """RPM package description (derived from package.description value from Cargo.toml)."""
        return self._conf.package.description or self._project.main_package.get_description()

    @property
    def rpm_url(self) -> str:
        """RPM URL tag (derived from the rust2rpm.toml setting or Cargo.toml metadata)."""
        return self._conf.package.url or "# FIXME"

    @property
    def rpm_source_url(self) -> str:
        """RPM Source tag (derive from the rust2rpm.toml setting)."""
        return self._conf.package.source_url or "# FIXME"

    @property
    def rpm_license(self) -> str | None:
        """RPM License tag (derived from package.license value from Cargo.toml)."""
        return self._rpm_license_and_comments[0]

    @property
    def rpm_license_comments(self) -> str | None:
        """Additional information returned by license string translation."""
        return self._rpm_license_and_comments[1]

    @property
    def rpm_patch_file_automatic(self) -> str | None:
        """File name of the automatically generated patch."""
        return None

    @property
    def rpm_patch_file_manual(self) -> str | None:
        """File name of the manually generated patch."""
        return None

    @property
    def rpm_patch_file_comments(self) -> list[str]:
        """Additional lines of comments for the manually generated patch file."""
        return self._conf.package.cargo_toml_patch_comment_lines

    @property
    def rpm_license_files(self) -> list[str]:
        """List of the license files which were detected in crate sources."""
        return self._project.license_files

    @property
    def rpm_doc_files(self) -> list[str]:
        """List of the documentation files which were detected in crate sources."""
        return self._project.doc_files

    @property
    def rpm_bcond_check(self) -> int:
        """Flag to switch default value of the "check" bcond."""
        return conf_to_bcond_check(self._conf)

    @property
    def rpm_bcond_check_comments(self) -> list[str]:
        """Comments associated with a disabled "check" bcond."""
        return conf_to_bcond_check_comments(self._conf)

    @property
    def rpm_vendor_source(self) -> str | None:
        """File name of the vendor tarball in case vendored sources are used."""
        return self._project.vendor_tarball

    @property
    def rpm_extra_sources(self) -> list[Source]:
        """Additional source files with number and comments."""
        return self._conf.package.extra_sources

    @property
    def rpm_extra_patches(self) -> list[Patch]:
        """Additional patch files with number and comments."""
        return self._conf.package.extra_patches

    @property
    def rpm_extra_files(self) -> list[str]:
        """Additional files to be included in the built package."""
        return self._conf.package.extra_files

    # Parameters that control generation of subpackages

    @property
    def rpm_binary_package(self) -> bool:
        """True if package ships any binary targets (bin targets)."""
        return self._metadata.is_bin()

    @property
    def rpm_cdylib_package(self) -> bool:
        """True if package ships any shared libraries (cdylib targets)."""
        return self._metadata.is_cdylib()

    @property
    def rpm_binary_names(self) -> list[str]:
        """List of the names of executables which are built from the crate."""
        # enforce consistent order
        if bin_renames := self._conf.package.bin_renames:
            binaries = []
            for old_name in self._metadata.get_binaries():
                if new_name := bin_renames.get(old_name):
                    binaries.append(new_name)
                else:
                    binaries.append(old_name)
        else:
            binaries = list(self._metadata.get_binaries())
        return sorted(binaries)

    # Parameters that allow injecting additional commands into scriptlets

    @property
    def rpm_prep_pre(self) -> list[str]:
        """Additional commands that are injected before %cargo_prep."""
        return self._conf.scripts.prep.pre

    @property
    def rpm_prep_post(self) -> list[str]:
        """Additional commands that are injected after %cargo_prep."""
        return self._conf.scripts.prep.post

    @property
    def rpm_build_pre(self) -> list[str]:
        """Additional commands that are injected before %cargo_build."""
        return self._conf.scripts.build.pre

    @property
    def rpm_build_post(self) -> list[str]:
        """Additional commands that are injected after %cargo_build."""
        return self._conf.scripts.build.post

    @property
    def rpm_install_pre(self) -> list[str]:
        """Additional commands that are injected before %cargo_install."""
        return self._conf.scripts.install.pre

    @property
    def rpm_install_post(self) -> list[str]:
        """Additional commands that are injected after %cargo_install."""
        return self._conf.scripts.install.post

    @property
    def rpm_check_pre(self) -> list[str]:
        """Additional commands that are injected before %cargo_check."""
        return self._conf.scripts.check.pre

    @property
    def rpm_check_post(self) -> list[str]:
        """Additional commands that are injected after %cargo_check."""
        return self._conf.scripts.check.post

    # Parameters for RPM macros

    @property
    def upstream_version(self) -> str | None:
        """Upstream project version."""
        return self._project.version

    @property
    def rpm_autosetup_args(self) -> str:
        """Additional arguments for the %autosetup macro."""
        return " -a1" if self._project.vendor_tarball else ""

    @property
    def cargo_args(self) -> str:
        """Additional arguments for %cargo_build, %cargo_install, etc."""
        feature_flags = conf_to_feature_flags(self._conf)
        return cargo_args_from_flags(feature_flags, set(), set())

    @property
    def cargo_prep_args(self) -> str:
        """Additional arguments for %cargo_prep."""
        return " -v vendor" if self._project.vendor_tarball else ""

    @property
    def cargo_test_commands(self) -> list[str]:
        """List of customized and pre-formatted %cargo_test commands."""
        return conf_to_cargo_test_commands(self._conf, self.cargo_args)

    # Parameters derived from rust2rpm.toml

    @property
    def conf_buildrequires(self) -> list[str]:
        """List of additionally specified RPM BuildRequires."""
        return self._conf.requires.build

    @property
    def conf_test_requires(self) -> list[str]:
        """List of additionally specified RPM BuildRequires that are gated by an "%if %{with check}"" conditional."""
        return self._conf.requires.test

    @property
    def conf_bin_requires(self) -> list[str]:
        """List of additionally specified RPM Requires for the binary package."""
        return self._conf.requires.bin

    @property
    def conf_supported_arches(self) -> str | None:
        """List of supported architectures.

        Results in an ExclusiveArch tag with the specified values.
        """
        if supported_arches := self._conf.package.supported_arches:
            conf_supported_arches = " ".join(supported_arches)
        else:
            conf_supported_arches = None

        return conf_supported_arches

    @property
    def conf_suppress_cdylib_install_fixme(self) -> bool:
        """Toggle suppression of the FIXME comment for installing cdylib crate targets."""
        return self._conf.package.suppress_cdylib_install_fixme is True

    @property
    def rpm_bin_renames(self) -> list[str]:
        """List of "mv" commands to rename binaries installed in %{_bindir}."""
        return conf_to_bin_rename_commands(self._conf)

    # Parameters derived from command-line flags

    @property
    def make_changelog_entry(self) -> bool:
        """Toggle inclusion of a changelog entry in generated spec files."""
        return self._make_changelog_entry


@dataclass(frozen=True)
class WorkspaceParametersFedora(WorkspaceParameters, ParametersFedora):
    """Collection of parameters for the "workspace" type template that is specific to the "fedora" target."""


@dataclass(frozen=True)
class WorkspaceParametersMageia(WorkspaceParameters, ParametersMageia):
    """Collection of parameters for the "workspace" type template that is specific to the "mageia" target."""

    @property
    def rpm_buildrequires(self) -> list[str]:
        """Automatically generated RPM BuildRequires for crate dependencies."""
        return self._buildrequires[0]

    @property
    def rpm_test_requires(self) -> list[str]:
        """Automatically generated test-only RPM BuildRequires for crate dependencies."""
        return self._buildrequires[1]


@dataclass(frozen=True)
class WorkspaceParametersOpenSUSE(WorkspaceParameters, ParametersOpenSUSE):
    """Collection of parameters for the "workspace" type template that is specific to the "opensuse" target."""

    @property
    def rpm_buildrequires(self) -> list[str]:
        """Automatically generated RPM BuildRequires for crate dependencies."""
        return self._buildrequires[0]

    @property
    def rpm_test_requires(self) -> list[str]:
        """Automatically generated test-only RPM BuildRequires for crate dependencies."""
        return self._buildrequires[1]


@dataclass(frozen=True)
class WorkspaceParametersPlain(WorkspaceParameters, ParametersPlain):
    """Collection of parameters for the "workspace" type template that is specific to the "plain" target."""

    @property
    def rpm_buildrequires(self) -> list[str]:
        """Automatically generated RPM BuildRequires for crate dependencies."""
        return self._buildrequires[0]

    @property
    def rpm_test_requires(self) -> list[str]:
        """Automatically generated test-only RPM BuildRequires for crate dependencies."""
        return self._buildrequires[1]

    @property
    def rpm_requires(self) -> dict[str | None, list[str]]:
        """Automatically generated RPM Requires for subpackages.

        This returns an empty dictionary because this value is unused in the "project" template.
        """
        return {}

    @property
    def rpm_provides(self) -> dict[str | None, str]:
        """Automatically generated RPM Provides for subpackages.

        This returns an empty dictionary because this value is unused in the "project" template.
        """
        return {}


@dataclass(frozen=True)
class WorkspaceTemplate(Template):
    """Template for rendering spec files with the "workspace" template."""

    project: LocalProject
    """Project for which the spec file is rendered."""

    conf: TomlConf
    """Global rust2rpm configuration."""

    target: Target
    """Target for which the spec file is rendered."""

    date: time.struct_time | None
    """Optional timestamp for generated changelog entries.
    Falls back to the current time if not specified."""

    packager: str | None
    """Optional user identification for generated changelog entries.
    Falls back to dummy values if not specified."""

    use_rpmautospec: bool
    """Predicate that controls whether rpmautospec is used for
    the Release tag and %changelog."""

    make_changelog_entry: bool
    """Predicate that controls whether a changelog entry is automatically generated."""

    def render(self) -> str:
        """Render spec file from template with the computed parameters."""
        template = spec_file_template("workspace.spec")

        match self.target:
            case Target.FEDORA:
                parameter_cls = WorkspaceParametersFedora  # type: ignore[assignment]
            case Target.MAGEIA:
                parameter_cls = WorkspaceParametersMageia  # type: ignore[assignment]
            case Target.OPENSUSE:
                parameter_cls = WorkspaceParametersOpenSUSE  # type: ignore[assignment]
            case Target.PLAIN:
                parameter_cls = WorkspaceParametersPlain  # type: ignore[assignment]
            case x:  # pragma nocover
                msg = f"{x} is not a supported target for workspace-type packages"
                raise ValueError(msg)

        parameters = parameter_cls(
            _project=self.project,
            _conf=self.conf,
            _target=self.target,
            _make_changelog_entry=self.make_changelog_entry,
            _date=self.date,
            _packager=self.packager,
            _use_rpmautospec=self.use_rpmautospec,
        )

        contents = template.render(**parameters_as_dict(parameters))

        if not contents.endswith("\n"):
            contents += "\n"

        return contents
