"""Module containing base project classes."""

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

from cargo2rpm.metadata import Metadata


class InvalidVersionError(ValueError):
    """Raised for version arguments that either don't parse as valid SemVer or don't resolve to an existing version."""


@dataclass(frozen=True)
class PatchFile:
    """Representation of an automatically generated or manually applied Patch for Cargo.toml."""

    name: str
    """File name of the patch."""

    contents: list[str]
    """Contents of the patch as a list of lines."""


@dataclass(frozen=True)
class Project(metaclass=ABCMeta):
    """Base class for loaded cargo projects."""

    name: str
    """Name of the project."""

    version: str
    """Version of the project."""

    metadata: Metadata
    """Metadata associated with the project (from `cargo metadata`)."""

    license_files: list[str]
    """List of license files found by crawling the project sources."""

    doc_files: list[str]
    """List of documentation files found by crawling the project sources."""

    auto_patch: PatchFile | None
    """Automatically generated patch for Cargo.toml (if present)."""

    manual_patch: PatchFile | None
    """Manually applied patch for Cargo.toml (if present)."""

    vendor_tarball: str | None
    """File name of the tarball that contains vendored dependencies (if present)."""

    @property
    @abstractmethod
    def rpm_name(self) -> str:
        """RPM source package name based on project metadata."""
